package be.iminds.ilabt.jfed.highlevel.jobs;

import be.iminds.ilabt.jfed.experiment.Experiment;
import be.iminds.ilabt.jfed.experiment.ExperimentPart;
import be.iminds.ilabt.jfed.experiment.tasks.ExperimentTaskStatus;
import be.iminds.ilabt.jfed.highlevel.controller.JFedHighLevelException;
import be.iminds.ilabt.jfed.highlevel.controller.TaskThread;
import be.iminds.ilabt.jfed.highlevel.jobs.parts.ExperimentPartController;
import be.iminds.ilabt.jfed.highlevel.jobs.parts.ExperimentPartControllerManager;
import be.iminds.ilabt.jfed.highlevel.tasks.HighLevelTaskFactory;
import be.iminds.ilabt.jfed.highlevel.tasks.RenewSliceTask;

import javax.annotation.Nonnull;
import javax.annotation.Nullable;
import java.time.Instant;
import java.time.ZoneId;
import java.time.ZonedDateTime;
import java.time.format.DateTimeFormatter;
import java.util.Collection;
import java.util.Collections;
import java.util.List;
import java.util.stream.Collectors;

/**
 * User: twalcari
 * Date: 12/2/2015
 * Time: 15:54
 */
public class RenewJob extends AbstractJob<ExperimentTaskStatus> {

    @Nonnull private final Instant newExpirationDate;
    @Nullable private final Collection<ExperimentPart> experimentParts; //null means ALL
    @Nullable private final boolean renewSliceOnly;

    @Nonnull private final ExperimentPartControllerManager epcManager;
    private List<ExperimentPart> failedParts;
    private boolean renewSliceFailed;

    public RenewJob(@Nonnull Experiment experiment,
                    @Nonnull Instant newExpirationDate,
                    @Nonnull HighLevelTaskFactory hltf,
                    @Nonnull TaskThread tt,
                    @Nonnull ExperimentPartControllerManager epcManager,
                    boolean renewSliceOnly) {
        super("Renew "+(renewSliceOnly ? "slice" : "experiment")+" until " + DateTimeFormatter.RFC_1123_DATE_TIME.format(ZonedDateTime.ofInstant(newExpirationDate, ZoneId.systemDefault())),
                experiment, hltf, tt);
        this.newExpirationDate = newExpirationDate;
        this.epcManager = epcManager;
        this.experimentParts = null;
        this.renewSliceOnly = renewSliceOnly;
    }

    public RenewJob(@Nonnull Experiment experiment,
                    @Nonnull Instant newExpirationDate,
                    @Nonnull Collection<ExperimentPart> experimentParts,
                    @Nonnull HighLevelTaskFactory hltf,
                    @Nonnull TaskThread tt,
                    @Nonnull ExperimentPartControllerManager epcManager) {
        super("Renew experiment until " + DateTimeFormatter.RFC_1123_DATE_TIME.format(ZonedDateTime.ofInstant(newExpirationDate, ZoneId.systemDefault())),
                experiment, hltf, tt);
        this.newExpirationDate = newExpirationDate;
        this.experimentParts = experimentParts;
        this.epcManager = epcManager;
        this.renewSliceOnly = false;
    }

    @Override
    public ExperimentTaskStatus execute() throws Exception {
        renewSliceFailed = false;
        assert experiment.getSliceOrNull() != null;
        if (experiment.getSlice().getExpirationDate() == null ||
                newExpirationDate.isAfter(experiment.getSlice().getExpirationDate())) {
            final RenewSliceState renewSliceState = new RenewSliceState();
            setAndRunState(renewSliceState);

            //don't throw an exception here: just try to renew the slivers anyway. We will throw the exception at
            //the end of the method
            renewSliceFailed =  (renewSliceState.getStatus() != ExperimentTaskStatus.SUCCESS);
        }

        if(renewSliceFailed)
            throw new JFedHighLevelException("Could not renew slice!");

        if (renewSliceOnly) {
            failedParts = Collections.emptyList();
            return ExperimentTaskStatus.SUCCESS;
        } else {
            RenewExperimentPartsState renewPartsState = new RenewExperimentPartsState();
            setAndRunState(renewPartsState);

            failedParts = renewPartsState.getSlices().stream()
                    .filter(js -> js.getResult() != ExperimentTaskStatus.SUCCESS)
                    .map(ExperimentPartStateSlice::getExperimentPart)
                    .collect(Collectors.toList());

            if (!failedParts.isEmpty()) {
                updateMessage("Renewing failed at " + failedParts.stream().map(ExperimentPart::getName).collect(Collectors.joining(", ")));
            }

            return renewPartsState.getStatus();
        }
    }

    private class RenewSliceState extends SingleCallState<RenewSliceTask> {

        protected RenewSliceState() {
            super("Renewing experiment container", hltf.renewSlice(experiment.getSlice(), newExpirationDate));
        }
    }

    private class RenewExperimentPartsState extends ExperimentPartsSlicedState<ExperimentPartStateSlice> {
        protected RenewExperimentPartsState() {
            super("Renewing resources at %d authorities", experiment, experimentParts, epcManager);
        }

        @Override
        public <EP extends ExperimentPart> ExperimentPartStateSlice createJobSlice(ExperimentPartController<EP> controller, EP experimentPart) {
            return controller.renew(RenewJob.this, experimentPart, newExpirationDate);
        }
    }

    public boolean isRenewSliceFailed() {
        return renewSliceFailed;
    }

    public List<ExperimentPart> getFailedParts() {
        return Collections.unmodifiableList(failedParts);
    }
}
